@@ -1,104 +1,22 @@
 import tensorflow as tf
-
+# import tensorflow.keras as ks
 from kgcnn.layers.base import GraphBaseLayer
 from kgcnn.layers.gather import GatherNodesIngoing, GatherNodesOutgoing, GatherState
-from kgcnn.layers.keras import Dense, Activation, Concatenate
+from kgcnn.layers.keras import Dense, Activation, Concatenate, Multiply, Add
 from kgcnn.layers.pooling import PoolingNodes
+from kgcnn.layers.pooling import PoolingNodesAttention, PoolingLocalEdgesAttention
 import kgcnn.ops.activ
-from kgcnn.ops.partition import kgcnn_ops_change_edge_tensor_indexing_by_row_partition
-from kgcnn.ops.scatter import kgcnn_ops_scatter_segment_tensor_nd
-from kgcnn.ops.segment import segment_softmax, kgcnn_ops_segment_operation_by_name
 
 
-# import tensorflow.keras as ks
-
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingLocalEdgesAttention')
-class PoolingLocalEdgesAttention(GraphBaseLayer):
-    r"""Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
-    Uses attention for pooling. i.e.  $n_i =  \sum_j \alpha_{ij} e_ij $
-    The attention is computed via: $\alpha_ij = softmax_j (a_ij)$ from the attention coefficients $a_ij$.
-    The attention coefficients must be computed beforehand by edge features or by $\sigma( W n_i || W n_j)$ and
-    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
-    In summary, $n_i =  \sum_j softmax_j(a_ij) e_ij $ is computed by the layer.
-
-    An edge is defined by index tuple (i,j) with i<-j connection.
-    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
-    Important: edge_index[:,0] are sorted for segment-operation for pooling with respect to edge_index[:,0].
-    """
-
-    def __init__(self, **kwargs):
-        """Initialize layer."""
-        super(PoolingLocalEdgesAttention, self).__init__(**kwargs)
-
-    def build(self, input_shape):
-        """Build layer."""
-        super(PoolingLocalEdgesAttention, self).build(input_shape)
-
-    def call(self, inputs, **kwargs):
-        """Forward pass.
-
-        Args:
-            inputs: [node, edges, attention, edge_indices]
-
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - edges (tf.ragged): Edge or message features of shape (batch, [M], F)
-            - attention (tf.ragged): Attention coefficients of shape (batch, [M], 1)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], F)
-
-        Returns:
-            embeddings: Feature tensor of pooled edge attentions for each node of shape (batch, [N], F)
-        """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 4)
-
-        # We cast to values here
-        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
-        edge = dyn_inputs[1].values
-        attention = dyn_inputs[2].values
-        edgeind, edge_part = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
-
-        shiftind = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edgeind, node_part, edge_part,
-                                                                          partition_type_node="row_length",
-                                                                          partition_type_edge="row_length",
-                                                                          to_indexing='batch',
-                                                                          from_indexing=self.node_indexing)
-
-        nodind = shiftind[:, 0]  # Pick first index eg. ingoing
-        dens = edge
-        ats = attention
-        if not self.is_sorted:
-            # Sort edgeindices
-            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
-            nodind = tf.gather(nodind, node_order, axis=0)
-            dens = tf.gather(dens, node_order, axis=0)
-            ats = tf.gather(ats, node_order, axis=0)
-
-        # Apply segmented softmax
-        ats = segment_softmax(ats, nodind)
-        get = dens * ats
-        get = tf.math.segment_sum(get, nodind)
-
-        if self.has_unconnected:
-            # Need to fill tensor since the maximum node may not be also in pooled
-            # Does not happen if all nodes are also connected
-            get = kgcnn_ops_scatter_segment_tensor_nd(get, nodind, tf.shape(nod))
-
-        out = self._kgcnn_map_output_ragged([get, node_part], "row_length", 0)
-        return out
-
-    def get_config(self):
-        """Update layer config."""
-        config = super(PoolingLocalEdgesAttention, self).get_config()
-        return config
-
-
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='AttentionHeadGAT')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentionHeadGAT')
 class AttentionHeadGAT(GraphBaseLayer):
-    r"""Computes the attention head according to GAT.
-    The attention coefficients are computed by $a_{ij} = \sigma( W n_i || W n_j)$,
-    optionally by $a_{ij} = \sigma( W n_i || W n_j || e_{ij})$.
-    The attention is obtained by $\alpha_ij = softmax_j (a_{ij})$.
-    And the messages are pooled by $n_i =  \sum_j \alpha_{ij} e_ij $.
-    And finally passed through an activation $h_i = \sigma(\sum_j \alpha_{ij} e_ij)$.
+    r"""Computes the attention head according to GAT from https://arxiv.org/abs/1710.10903.
+    The attention coefficients are computed by :math:`a_{ij} = \sigma(a^T W n_i || W n_j)`,
+    optionally by :math:`a_{ij} = \sigma( W n_i || W n_j || e_{ij})` with edges :math:`e_{ij}`.
+    The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
+    And the messages are pooled by :math:`m_i = \sum_j \alpha_{ij} W n_j`.
+    If the graph has no self-loops, they must be added beforehand or use external skip connections.
+    And optionally passed through an activation :math:`h_i = \sigma(\sum_j \alpha_{ij} W n_j)`.
 
     An edge is defined by index tuple (i,j) with i<-j connection.
     If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
@@ -106,6 +24,8 @@ class AttentionHeadGAT(GraphBaseLayer):
     Args:
         units (int): Units for the linear trafo of node features before attention.
         use_edge_features (bool): Append edge features to attention computation. Default is False.
+        use_final_activation (bool): Whether to apply the final activation for the output.
+        has_self_loops (bool): If the graph has self-loops. Not used here. Default is True.
         activation (str): Activation. Default is {"class_name": "kgcnn>leaky_relu", "config": {"alpha": 0.2}},
         use_bias (bool): Use bias. Default is True.
         kernel_regularizer: Kernel regularization. Default is None.
@@ -120,6 +40,8 @@ class AttentionHeadGAT(GraphBaseLayer):
     def __init__(self,
                  units,
                  use_edge_features=False,
+                 use_final_activation=True,
+                 has_self_loops=True,
                  activation='kgcnn>leaky_relu',
                  use_bias=True,
                  kernel_regularizer=None,
@@ -134,17 +56,20 @@ class AttentionHeadGAT(GraphBaseLayer):
         super(AttentionHeadGAT, self).__init__(**kwargs)
         # graph args
         self.use_edge_features = use_edge_features
+        self.use_final_activation = use_final_activation
+        self.has_self_loops = has_self_loops
 
         # dense args
         self.units = int(units)
+        self.use_bias = use_bias
 
-        kernel_args = {"use_bias": use_bias, "kernel_regularizer": kernel_regularizer,
+        kernel_args = {"kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", **kernel_args, **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation=activation, **kernel_args, **self._kgcnn_info)
+        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
+        self.lay_alpha = Dense(1, activation=activation, use_bias=False, **kernel_args, **self._kgcnn_info)
         self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
         self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
         self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
@@ -161,47 +86,167 @@ class AttentionHeadGAT(GraphBaseLayer):
         Args:
             inputs (list): of [node, edges, edge_indices]
 
-            - nodes: Node features of shape (batch, [N], F)
-            - edges: Edge or message features of shape (batch, [M], F)
-            - edge_index: Edge indices of shape (batch, [M], 2)
+                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
+                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
 
         Returns:
-            features: Feature tensor of pooled edge attentions for each node.
+            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node.
         """
         node, edge, edge_index = inputs
 
-        n_in = self.lay_gather_in([node, edge_index])
-        n_out = self.lay_gather_out([node, edge_index])
-        wn_in = self.lay_linear_trafo(n_in)
-        wn_out = self.lay_linear_trafo(n_out)
+        w_n = self.lay_linear_trafo(node)
+        wn_in = self.lay_gather_in([w_n, edge_index])
+        wn_out = self.lay_gather_out([w_n, edge_index])
         if self.use_edge_features:
             e_ij = self.lay_concat([wn_in, wn_out, edge])
         else:
             e_ij = self.lay_concat([wn_in, wn_out])
         a_ij = self.lay_alpha(e_ij)  # Should be dimension (batch*None,1)
-        n_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
-        out = self.lay_final_activ(n_i)
-        return out
+        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
+
+        if self.use_final_activation:
+            h_i = self.lay_final_activ(h_i)
+        return h_i
 
     def get_config(self):
         """Update layer config."""
         config = super(AttentionHeadGAT, self).get_config()
-        config.update({"use_edge_features": self.use_edge_features,
-                       "units": self.units})
+        config.update({"use_edge_features": self.use_edge_features, "use_bias": self.use_bias,
+                       "units": self.units, "has_self_loops": self.has_self_loops,
+                       "use_final_activation": self.use_final_activation})
         conf_sub = self.lay_alpha.get_config()
         for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
-                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
+                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
             config.update({x: conf_sub[x]})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='AttentiveHeadFP')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentionHeadGATV2')
+class AttentionHeadGATV2(GraphBaseLayer):
+    r"""Computes the modified attention head according to https://arxiv.org/pdf/2105.14491.pdf.
+    The attention coefficients are computed by :math:`a_{ij} = a^T \sigma( W [n_i || n_j] )`,
+    optionally by :math:`a_{ij} = a^T \sigma( W [n_i || n_j || e_{ij}] )` with edges :math:`e_{ij}`.
+    The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
+    And the messages are pooled by :math:`m_i =  \sum_j \alpha_{ij} e_{ij}`.
+    If the graph has no self-loops, they must be added beforehand or use external skip connections.
+    And optionally passed through an activation :math:`h_i = \sigma(\sum_j \alpha_{ij} e_{ij})`.
+
+    An edge is defined by index tuple (i,j) with i<-j connection.
+    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+
+    Args:
+        units (int): Units for the linear trafo of node features before attention.
+        use_edge_features (bool): Append edge features to attention computation. Default is False.
+        use_final_activation (bool): Whether to apply the final activation for the output.
+        has_self_loops (bool): If the graph has self-loops. Not used here. Default is True.
+        activation (str): Activation. Default is {"class_name": "kgcnn>leaky_relu", "config": {"alpha": 0.2}},
+        use_bias (bool): Use bias. Default is True.
+        kernel_regularizer: Kernel regularization. Default is None.
+        bias_regularizer: Bias regularization. Default is None.
+        activity_regularizer: Activity regularization. Default is None.
+        kernel_constraint: Kernel constrains. Default is None.
+        bias_constraint: Bias constrains. Default is None.
+        kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
+        bias_initializer: Initializer for bias. Default is 'zeros'.
+    """
+
+    def __init__(self,
+                 units,
+                 use_edge_features=False,
+                 use_final_activation=True,
+                 has_self_loops=True,
+                 activation='kgcnn>leaky_relu',
+                 use_bias=True,
+                 kernel_regularizer=None,
+                 bias_regularizer=None,
+                 activity_regularizer=None,
+                 kernel_constraint=None,
+                 bias_constraint=None,
+                 kernel_initializer='glorot_uniform',
+                 bias_initializer='zeros',
+                 **kwargs):
+        """Initialize layer."""
+        super(AttentionHeadGATV2, self).__init__(**kwargs)
+        # graph args
+        self.use_edge_features = use_edge_features
+        self.use_final_activation = use_final_activation
+        self.has_self_loops = has_self_loops
+
+        # dense args
+        self.units = int(units)
+        self.use_bias = use_bias
+
+        kernel_args = {"kernel_regularizer": kernel_regularizer,
+                       "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
+                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
+                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
+
+        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
+        self.lay_alpha_activation = Dense(units, activation=activation, use_bias=use_bias, **kernel_args,
+                                          **self._kgcnn_info)
+        self.lay_alpha = Dense(1, activation="linear", use_bias=False, **kernel_args, **self._kgcnn_info)
+        self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
+        self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
+        self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
+        self.lay_pool_attention = PoolingLocalEdgesAttention(**self._kgcnn_info)
+        self.lay_final_activ = Activation(activation=activation, **self._kgcnn_info)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(AttentionHeadGATV2, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs (list): of [node, edges, edge_indices]
+
+                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
+                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
+
+        Returns:
+            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node.
+        """
+        node, edge, edge_index = inputs
+
+        w_n = self.lay_linear_trafo(node)
+        n_in = self.lay_gather_in([node, edge_index])
+        n_out = self.lay_gather_out([node, edge_index])
+        wn_out = self.lay_gather_out([w_n, edge_index])
+        if self.use_edge_features:
+            e_ij = self.lay_concat([n_in, n_out, edge])
+        else:
+            e_ij = self.lay_concat([n_in, n_out])
+        a_ij = self.lay_alpha_activation(e_ij)
+        a_ij = self.lay_alpha(a_ij)
+        h_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
+
+        if self.use_final_activation:
+            h_i = self.lay_final_activ(h_i)
+        return h_i
+
+    def get_config(self):
+        """Update layer config."""
+        config = super(AttentionHeadGATV2, self).get_config()
+        config.update({"use_edge_features": self.use_edge_features, "use_bias": self.use_bias,
+                       "units": self.units, "has_self_loops": self.has_self_loops,
+                       "use_final_activation": self.use_final_activation})
+        conf_sub = self.lay_alpha_activation.get_config()
+        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
+                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
+            config.update({x: conf_sub[x]})
+        return config
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='AttentiveHeadFP')
 class AttentiveHeadFP(GraphBaseLayer):
     r"""Computes the attention head for Attentive FP model.
-    The attention coefficients are computed by $a_{ij} = \sigma_1( W_1 [h_i || h_j] )$.
-    The initial representation $h_i$ and $h_j$ must be calculated beforehand.
-    The attention is obtained by $\alpha_ij = softmax_j (a_{ij})$.
-    And finally pooled through for context $C_i = \sigma_2(\sum_j \alpha_{ij} W_2 h_j)$.
+    The attention coefficients are computed by :math:`a_{ij} = \sigma_1( W_1 [h_i || h_j] )`.
+    The initial representation :math:`h_i` and :math:`h_j` must be calculated beforehand.
+    The attention is obtained by :math:`\alpha_{ij} = \text{softmax}_j (a_{ij})`.
+    And finally pooled for context :math:`C_i = \sigma_2(\sum_j \alpha_{ij} W_2 h_j)`.
 
     If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
 
@@ -241,22 +286,25 @@ class AttentiveHeadFP(GraphBaseLayer):
 
         # dense args
         self.units = int(units)
+        self.use_bias = use_bias
 
-        kernel_args = {"use_bias": use_bias, "kernel_regularizer": kernel_regularizer,
+        kernel_args = {"kernel_regularizer": kernel_regularizer,
                        "activity_regularizer": activity_regularizer, "bias_regularizer": bias_regularizer,
                        "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                        "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}
 
-        self.lay_linear_trafo = Dense(units, activation="linear", **kernel_args, **self._kgcnn_info)
-        self.lay_alpha = Dense(1, activation=activation, **kernel_args, **self._kgcnn_info)
+        self.lay_linear_trafo = Dense(units, activation="linear", use_bias=use_bias, **kernel_args, **self._kgcnn_info)
+        self.lay_alpha_activation = Dense(units, activation=activation, use_bias=use_bias, **kernel_args,
+                                          **self._kgcnn_info)
+        self.lay_alpha = Dense(1, activation="linear", use_bias=False, **kernel_args, **self._kgcnn_info)
         self.lay_gather_in = GatherNodesIngoing(**self._kgcnn_info)
         self.lay_gather_out = GatherNodesOutgoing(**self._kgcnn_info)
         self.lay_concat = Concatenate(axis=-1, **self._kgcnn_info)
         self.lay_pool_attention = PoolingLocalEdgesAttention(**self._kgcnn_info)
         self.lay_final_activ = Activation(activation=activation_context, **self._kgcnn_info)
         if use_edge_features:
-            self.lay_fc1 = Dense(units, activation=activation, **kernel_args, **self._kgcnn_info)
-            self.lay_fc2 = Dense(units, activation=activation, **kernel_args, **self._kgcnn_info)
+            self.lay_fc1 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args, **self._kgcnn_info)
+            self.lay_fc2 = Dense(units, activation=activation, use_bias=use_bias, **kernel_args, **self._kgcnn_info)
             self.lay_concat_edge = Concatenate(axis=-1, **self._kgcnn_info)
 
     def build(self, input_shape):
@@ -269,12 +317,12 @@ class AttentiveHeadFP(GraphBaseLayer):
         Args:
             inputs (list): of [node, edges, edge_indices]
 
-            - nodes: Node features of shape (batch, [N], F)
-            - edges: Edge or message features of shape (batch, [M], F)
-            - edge_index: Edge indices of shape (batch, [M], 2)
+                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
+                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
 
         Returns:
-            features: Hidden tensor of pooled edge attentions for each node.
+            tf.RaggedTensor: Hidden tensor of pooled edge attentions for each node.
         """
         node, edge, edge_index = inputs
 
@@ -290,7 +338,9 @@ class AttentiveHeadFP(GraphBaseLayer):
 
         wn_out = self.lay_linear_trafo(n_out)
         e_ij = self.lay_concat([n_in, n_out])
-        a_ij = self.lay_alpha(e_ij)  # Should be dimension (batch,None,1)
+        e_ij = self.lay_alpha_activation(e_ij)  # Maybe uses GAT original definition.
+        # a_ij = e_ij
+        a_ij = self.lay_alpha(e_ij)  # Should be dimension (batch,None,1) not fully clear in original paper SI.
         n_i = self.lay_pool_attention([node, wn_out, a_ij, edge_index])
         out = self.lay_final_activ(n_i)
         return out
@@ -298,67 +348,19 @@ class AttentiveHeadFP(GraphBaseLayer):
     def get_config(self):
         """Update layer config."""
         config = super(AttentiveHeadFP, self).get_config()
-        config.update({"use_edge_features": self.use_edge_features,
+        config.update({"use_edge_features": self.use_edge_features, "use_bias": self.use_bias,
                        "units": self.units})
-        conf_sub = self.lay_alpha.get_config()
+        conf_sub = self.lay_alpha_activation.get_config()
         for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
-                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
+                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
             config.update({x: conf_sub[x]})
         conf_context = self.lay_final_activ.get_config()
         config.update({"activation_context": conf_context["activation"]})
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingNodesAttention')
-class PoolingNodesAttention(GraphBaseLayer):
-    r"""Pooling all nodes
-    Uses attention for pooling. i.e.  $s =  \sum_j \alpha_{i} n_i $
-    The attention is computed via: $\alpha_i = softmax_i (a_i)$ from the attention coefficients $a_i$.
-    The attention coefficients must be computed beforehand by edge features or by $\sigma( W [s || n_i])$ and
-    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
-    In summary, $s =  \sum_i softmax_j(a_i) n_i $ is computed by the layer.
-
-    """
-
-    def __init__(self, **kwargs):
-        """Initialize layer."""
-        super(PoolingNodesAttention, self).__init__(**kwargs)
-
-    def build(self, input_shape):
-        """Build layer."""
-        super(PoolingNodesAttention, self).build(input_shape)
-
-    def call(self, inputs, **kwargs):
-        """Forward pass.
-
-        Args:
-            inputs: [nodes, attention]
-
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - attention (tf.ragged): Attention coefficients of shape (batch, [N], 1)
-
-        Returns:
-            embeddings: Feature tensor of pooled node of shape (batch, F)
-        """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 2)
-        # We cast to values here
-        nod, batchi, target_len = dyn_inputs[0].values, dyn_inputs[0].value_rowids(), dyn_inputs[0].row_lengths()
-        ats = dyn_inputs[1].values
-
-        ats = segment_softmax(ats, batchi)
-        get = nod * ats
-        out = tf.math.segment_sum(get, batchi)
-
-        return out
-
-    def get_config(self):
-        """Update layer config."""
-        config = super(PoolingNodesAttention, self).get_config()
-        return config
-
-
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='AttentiveNodePooling')
-class AttentiveNodePooling(GraphBaseLayer):
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodesAttentive')
+class PoolingNodesAttentive(GraphBaseLayer):
     r"""Computes the attentive pooling for node embeddings.
 
     Args:
@@ -400,7 +402,7 @@ class AttentiveNodePooling(GraphBaseLayer):
                  reset_after=True,
                  **kwargs):
         """Initialize layer."""
-        super(AttentiveNodePooling, self).__init__(**kwargs)
+        super(PoolingNodesAttentive, self).__init__(**kwargs)
         self.pooling_method = pooling_method
         self.depth = depth
         # dense args
@@ -429,7 +431,7 @@ class AttentiveNodePooling(GraphBaseLayer):
 
     def build(self, input_shape):
         """Build layer."""
-        super(AttentiveNodePooling, self).build(input_shape)
+        super(PoolingNodesAttentive, self).build(input_shape)
 
     def call(self, inputs, **kwargs):
         """Forward pass.
@@ -437,29 +439,29 @@ class AttentiveNodePooling(GraphBaseLayer):
         Args:
             inputs: nodes
 
-            - nodes: Node features of shape (batch, [N], F)
+                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
 
         Returns:
-            features: Hidden tensor of pooled edge attentions for each node.
+            tf.Tensor: Hidden tensor of pooled node attentions of shape (batch, F).
         """
         node = inputs
 
         h = self.lay_pool_start(node)
-        Wn = self.lay_linear_trafo(node)
+        wn = self.lay_linear_trafo(node)
         for _ in range(self.depth):
             hv = self.lay_gather_s([h, node])
             ev = self.lay_concat([hv, node])
             av = self.lay_alpha(ev)
-            cont = self.lay_pool_attention([Wn, av])
+            cont = self.lay_pool_attention([wn, av])
             cont = self.lay_final_activ(cont)
-            h,_ = self.lay_gru(cont, h, **kwargs)
+            h, _ = self.lay_gru(cont, h, **kwargs)
 
         out = h
         return out
 
     def get_config(self):
         """Update layer config."""
-        config = super(AttentiveNodePooling, self).get_config()
+        config = super(PoolingNodesAttentive, self).get_config()
         config.update({"units": self.units, "depth": self.depth, "pooling_method": self.pooling_method})
         conf_sub = self.lay_alpha.get_config()
         for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
@@ -468,7 +470,7 @@ class AttentiveNodePooling(GraphBaseLayer):
         conf_context = self.lay_final_activ.get_config()
         config.update({"activation_context": conf_context["activation"]})
         conf_gru = self.lay_gru.get_config()
-        for x in ["recurrent_activation", "recurrent_initializer", "mrecurrent_regularizer", "recurrent_constraint",
-                 "dropout", "recurrent_dropout", "reset_after"]:
+        for x in ["recurrent_activation", "recurrent_initializer", "recurrent_regularizer", "recurrent_constraint",
+                  "dropout", "recurrent_dropout", "reset_after"]:
             config.update({x: conf_gru[x]})
         return config
