@@ -2,12 +2,12 @@ import numpy as np
 import tensorflow as tf
 
 from kgcnn.layers.base import GraphBaseLayer
-from kgcnn.ops.partition import kgcnn_ops_change_edge_tensor_indexing_by_row_partition
+from kgcnn.ops.partition import partition_row_indexing
 from kgcnn.ops.polynom import spherical_bessel_jn_zeros, spherical_bessel_jn_normalization_prefactor, \
     tf_spherical_bessel_jn, tf_spherical_harmonics_yl
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='NodeDistance')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='NodeDistance')
 class NodeDistance(GraphBaseLayer):
     """Compute geometric node distances similar to edges.
 
@@ -29,29 +29,28 @@ class NodeDistance(GraphBaseLayer):
         Args:
             inputs (list): [position, edge_index]
 
-            - position (tf.ragged): Node positions of shape (batch, [N], 3)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], 2)
+                - position (tf.RaggedTensor): Node positions of shape (batch, [N], 3)
+                - edge_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
 
         Returns:
-            distances: Gathered node distances as edges that match the number of indices of shape (batch, [M], 1)
+            tf.RaggedTensor: Gathered node distances as edges that match the number of indices of shape (batch, [M], 1)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 2)
+        dyn_inputs = inputs
         # We cast to values here
         node, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge_index, edge_part = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
 
-        indexlist = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edge_index, node_part, edge_part,
-                                                                           partition_type_node="row_splits",
-                                                                           partition_type_edge="row_length",
-                                                                           to_indexing='batch',
-                                                                           from_indexing=self.node_indexing)
+        indexlist = partition_row_indexing(edge_index, node_part, edge_part, partition_type_target="row_splits",
+                                           partition_type_index="row_length", to_indexing='batch',
+                                           from_indexing=self.node_indexing)
         # For ragged tensor we can now also try:
-        # out = tf.gather(nod, edge_index[:, :, 0], batch_dims=1)
+        # out = tf.gather(nod, tensor_index[:, :, 0], batch_dims=1)
         xi = tf.gather(node, indexlist[:, 0], axis=0)
         xj = tf.gather(node, indexlist[:, 1], axis=0)
 
         out = tf.expand_dims(tf.sqrt(tf.nn.relu(tf.reduce_sum(tf.math.square(xi - xj), axis=-1))), axis=-1)
-        out = self._kgcnn_map_output_ragged([out, edge_part], "row_length", 1)
+
+        out = tf.RaggedTensor.from_row_lengths(out, edge_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -60,7 +59,145 @@ class NodeDistance(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='NodeAngle')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ScalarProduct')
+class ScalarProduct(GraphBaseLayer):
+    """Compute geometric scalar product for edges.
+
+    A distance based edge or node coordinates are defined by (batch, [N], ..., D) with last dimension D.
+    """
+
+    def __init__(self, **kwargs):
+        """Initialize layer."""
+        super(ScalarProduct, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(ScalarProduct, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs (list): [vec1, vec1]
+
+                - vec1 (tf.RaggedTensor): Positions of shape (batch, [N], ..., D)
+                - vec1 (tf.RaggedTensor): Positions of shape (batch, [N], ..., D)
+
+        Returns:
+            tf.RaggedTensor: Scalar product of shape (batch, [N], ...)
+        """
+        if all([isinstance(x, tf.RaggedTensor) for x in inputs]) and all([x.ragged_rank == 1 for x in inputs]):
+            v1 = inputs[0].values
+            v2 = inputs[1].values
+            out = tf.reduce_sum(v1*v2, axis=-1)
+            out = tf.RaggedTensor.from_row_splits(out, inputs[0].row_splits, validate=self.ragged_validate)
+        else:
+            out = tf.reduce_sum(inputs[0]*inputs[1], axis=-1)
+        return out
+
+    def get_config(self):
+        """Update config."""
+        config = super(ScalarProduct, self).get_config()
+        return config
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EuclideanNorm')
+class EuclideanNorm(GraphBaseLayer):
+    """Compute geometric norm for edges or nodes.
+
+    A distance based edge or node coordinates are defined by (batch, [N], ..., D) with last dimension D.
+    """
+
+    def __init__(self, **kwargs):
+        """Initialize layer."""
+        super(EuclideanNorm, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(EuclideanNorm, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs (list): coord
+
+                - coord (tf.RaggedTensor): Positions of shape (batch, [N], ..., D)
+
+        Returns:
+            tf.RaggedTensor: Scalar product of shape (batch, [N], ...)
+        """
+        # axis = [i for i in range(len(inputs.shape))][self.axis]
+        if isinstance(inputs, tf.RaggedTensor) and inputs.ragged_rank == 1:
+            v = inputs.values
+            out = tf.reduce_sum(tf.square(v), axis=-1)
+            out = tf.RaggedTensor.from_row_splits(out, inputs.row_splits, validate=self.ragged_validate)
+        else:  # Try general tensor approach
+            out = tf.reduce_sum(tf.square(inputs), axis=-1)
+        return out
+
+    def get_config(self):
+        """Update config."""
+        config = super(EuclideanNorm, self).get_config()
+        return config
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EdgeDirectionNormalized')
+class EdgeDirectionNormalized(GraphBaseLayer):
+    r"""Compute the normalized geometric edge direction similar to edges.
+    Will return :math:`\frac{r_{ij}}{||{r_ij}||} `.
+
+    A distance based edge is defined by edge or bond index in index list of shape (batch, [N], 2) with last dimension
+    of ingoing and outgoing.
+    """
+
+    def __init__(self, **kwargs):
+        """Initialize layer."""
+        super(EdgeDirectionNormalized, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(EdgeDirectionNormalized, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs (list): [position, edge_index]
+
+                - position (tf.RaggedTensor): Node positions of shape (batch, [N], 3)
+                - edge_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
+
+        Returns:
+            tf.RaggedTensor: Gathered node distances as edges that match the number of indices of shape (batch, [M], 3)
+        """
+        assert all([isinstance(x, tf.RaggedTensor) for x in inputs]) and all([x.ragged_rank == 1 for x in inputs])
+        dyn_inputs = inputs
+        # We cast to values here
+        node, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
+        edge_index, edge_part = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
+
+        indexlist = partition_row_indexing(edge_index, node_part, edge_part, partition_type_target="row_splits",
+                                           partition_type_index="row_length", to_indexing='batch',
+                                           from_indexing=self.node_indexing)
+        # For ragged tensor we can now also try:
+        # out = tf.gather(nod, tensor_index[:, :, 0], batch_dims=1)
+        xi = tf.gather(node, indexlist[:, 0], axis=0)
+        xj = tf.gather(node, indexlist[:, 1], axis=0)
+
+        xij = xi - xj
+        out = tf.expand_dims(tf.sqrt(tf.nn.relu(tf.reduce_sum(tf.math.square(xij), axis=-1))), axis=-1)
+        out = xij/out
+        out = tf.RaggedTensor.from_row_lengths(out, edge_part, validate=self.ragged_validate)
+        return out
+
+    def get_config(self):
+        """Update config."""
+        config = super(EdgeDirectionNormalized, self).get_config()
+        return config
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='NodeAngle')
 class NodeAngle(GraphBaseLayer):
     """Compute geometric node angles.
 
@@ -79,26 +216,24 @@ class NodeAngle(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (list): [position, edge_index]
+            inputs (list): [position, node_index]
 
-            - position (tf.ragged): Node positions of shape (batch, [N], 3)
-            - node_index (tf.ragged): Node indices of shape (batch, [M], 3) referring to nodes
+                - position (tf.RaggedTensor): Node positions of shape (batch, [N], 3)
+                - node_index (tf.RaggedTensor): Node indices of shape (batch, [M], 3) referring to nodes
 
         Returns:
-            angles: Gathered node angles between edges that match the indices. Shape is (batch, [M], 1)
+            tf.RaggedTensor: Gathered node angles between edges that match the indices. Shape is (batch, [M], 1)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 2)
+        dyn_inputs = inputs
         # We cast to values here
         node, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge_index, edge_part = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
 
-        indexlist = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edge_index, node_part, edge_part,
-                                                                           partition_type_node="row_splits",
-                                                                           partition_type_edge="row_length",
-                                                                           to_indexing='batch',
-                                                                           from_indexing=self.node_indexing)
+        indexlist = partition_row_indexing(edge_index, node_part, edge_part, partition_type_target="row_splits",
+                                           partition_type_index="row_length", to_indexing='batch',
+                                           from_indexing=self.node_indexing)
         # For ragged tensor we can now also try:
-        # out = tf.gather(nod, edge_index[:, :, 0], batch_dims=1)
+        # out = tf.gather(nod, tensor_index[:, :, 0], batch_dims=1)
         xi = tf.gather(node, indexlist[:, 0], axis=0)
         xj = tf.gather(node, indexlist[:, 1], axis=0)
         xk = tf.gather(node, indexlist[:, 2], axis=0)
@@ -110,7 +245,7 @@ class NodeAngle(GraphBaseLayer):
         angle = tf.math.atan2(y, x)
         angle = tf.expand_dims(angle, axis=-1)
 
-        out = self._kgcnn_map_output_ragged([angle, edge_part], "row_length", 1)
+        out = tf.RaggedTensor.from_row_lengths(angle, edge_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -119,7 +254,7 @@ class NodeAngle(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='EdgeAngle')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='EdgeAngle')
 class EdgeAngle(GraphBaseLayer):
     """Compute geometric edge angles.
 
@@ -138,34 +273,30 @@ class EdgeAngle(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (list): [position, edge_index]
+            inputs (list): [position, edge_index, angle_index]
 
-            - position (tf.ragged): Node positions of shape (batch, [N], 3)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], 2) referring to nodes.
-            - angle_index (tf.ragged): Angle indices of shape (batch, [K], 2) referring to edges.
+                - position (tf.RaggedTensor): Node positions of shape (batch, [N], 3)
+                - edge_index (tf.RaggedTensor): Edge indices of shape (batch, [M], 2) referring to nodes.
+                - angle_index (tf.RaggedTensor): Angle indices of shape (batch, [K], 2) referring to edges.
 
         Returns:
-            angles: Gathered edge angles between edges that match the indices. Shape is (batch, [K], 1)
+            tf.RaggedTensor: Gathered edge angles between edges that match the indices. Shape is (batch, [K], 1)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 3)
+        dyn_inputs = inputs
         node, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge_index, edge_part = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
         angle_index, angle_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
 
-        indexlist = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edge_index, node_part, edge_part,
-                                                                           partition_type_node="row_splits",
-                                                                           partition_type_edge="row_length",
-                                                                           to_indexing='batch',
-                                                                           from_indexing=self.node_indexing)
+        indexlist = partition_row_indexing(edge_index, node_part, edge_part, partition_type_target="row_splits",
+                                           partition_type_index="row_length", to_indexing='batch',
+                                           from_indexing=self.node_indexing)
 
-        indexlist2 = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(angle_index, edge_part, angle_part,
-                                                                            partition_type_node="row_splits",
-                                                                            partition_type_edge="row_length",
-                                                                            to_indexing='batch',
-                                                                            from_indexing=self.node_indexing)
+        indexlist2 = partition_row_indexing(angle_index, edge_part, angle_part, partition_type_target="row_splits",
+                                            partition_type_index="row_length", to_indexing='batch',
+                                            from_indexing=self.node_indexing)
 
         # For ragged tensor we can now also try:
-        # out = tf.gather(nod, edge_index[:, :, 0], batch_dims=1)
+        # out = tf.gather(nod, tensor_index[:, :, 0], batch_dims=1)
         xi = tf.gather(node, indexlist[:, 0], axis=0)
         xj = tf.gather(node, indexlist[:, 1], axis=0)
         vs = xi - xj
@@ -177,7 +308,7 @@ class EdgeAngle(GraphBaseLayer):
         angle = tf.math.atan2(y, x)
         angle = tf.expand_dims(angle, axis=-1)
 
-        out = self._kgcnn_map_output_ragged([angle, angle_part], "row_length", 2)
+        out = tf.RaggedTensor.from_row_lengths(angle, angle_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -186,10 +317,10 @@ class EdgeAngle(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='BesselBasisLayer')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='BesselBasisLayer')
 class BesselBasisLayer(GraphBaseLayer):
-    """
-    Expand a distance into a Bessel Basis with l=m=0, according to Klicpera et al. 2020
+    r"""
+    Expand a distance into a Bessel Basis with :math:`l=m=0`, according to Klicpera et al. 2020
 
     Args:
         num_radial (int): Number of radial radial basis functions
@@ -230,19 +361,19 @@ class BesselBasisLayer(GraphBaseLayer):
         Args:
             inputs: distance
 
-            - distance (tf.ragged): Edge distance of shape (batch, [K], 1)
+                - distance (tf.RaggedTensor): Edge distance of shape (batch, [K], 1)
 
         Returns:
-            distances: Expanded distance. Shape is (batch, [K], #Radial)
+            tf.RaggedTensor: Expanded distance. Shape is (batch, [K], #Radial)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged([inputs], 1)
+        dyn_inputs = [inputs]
         # We cast to values here
         node, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
 
         d_scaled = node * self.inv_cutoff
         d_cutoff = self.envelope(d_scaled)
         out = d_cutoff * tf.sin(self.frequencies * d_scaled)
-        out = self._kgcnn_map_output_ragged([out, node_part], "row_splits", 0)
+        out = tf.RaggedTensor.from_row_splits(out, node_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -253,10 +384,9 @@ class BesselBasisLayer(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='SphericalBasisLayer')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='SphericalBasisLayer')
 class SphericalBasisLayer(GraphBaseLayer):
-    """
-    Expand a distance into a Bessel Basis with l=m=0, according to Klicpera et al. 2020
+    r"""Expand a distance into a Bessel Basis with :math:`l=m=0`, according to Klicpera et al. 2020
 
     Args:
         num_spherical (int): Number of spherical basis functions
@@ -298,23 +428,21 @@ class SphericalBasisLayer(GraphBaseLayer):
         Args:
             inputs: [distance, angles, angle_index]
 
-            - distance (tf.ragged): Edge distance of shape (batch, [M], 1)
-            - angles (tf.ragged): Angle list of shape (batch, [K], 1)
-            - angle_index (tf.ragged): Indices referring to edges of shape (batch, [K], 2)
+                - distance (tf.RaggedTensor): Edge distance of shape (batch, [M], 1)
+                - angles (tf.RaggedTensor): Angle list of shape (batch, [K], 1)
+                - angle_index (tf.RaggedTensor): Angle indices referring to edges of shape (batch, [K], 2)
 
         Returns:
-            angles: Expanded angle/distance basis. Shape is (batch, [K], #Radial * #Spherical)
+            tf.RaggedTensor: Expanded angle/distance basis. Shape is (batch, [K], #Radial * #Spherical)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 3)
+        dyn_inputs = inputs
         edge, edge_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         angles, angle_part = dyn_inputs[1].values, dyn_inputs[1].row_splits
         angle_index, angle_index_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
 
-        indexlist = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(angle_index, edge_part, angle_index_part,
-                                                                           partition_type_node="row_splits",
-                                                                           partition_type_edge="row_length",
-                                                                           to_indexing='batch',
-                                                                           from_indexing=self.node_indexing)
+        indexlist = partition_row_indexing(angle_index, edge_part, angle_index_part,
+                                           partition_type_target="row_splits", partition_type_index="row_length",
+                                           to_indexing='batch', from_indexing=self.node_indexing)
 
         d = edge
         id_expand_kj = indexlist
@@ -335,7 +463,7 @@ class SphericalBasisLayer(GraphBaseLayer):
         cbf = tf.repeat(cbf, self.num_radial, axis=1)
         out = rbf_env * cbf
 
-        out = self._kgcnn_map_output_ragged([out, angle_part], "row_splits", 0) # out
+        out = tf.RaggedTensor.from_row_splits(out, angle_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
