@@ -2,18 +2,17 @@ import tensorflow as tf
 import tensorflow.keras as ks
 
 from kgcnn.layers.base import GraphBaseLayer
-from kgcnn.ops.partition import kgcnn_ops_change_edge_tensor_indexing_by_row_partition
-from kgcnn.ops.scatter import kgcnn_ops_scatter_segment_tensor_nd
-from kgcnn.ops.segment import kgcnn_ops_segment_operation_by_name
+from kgcnn.ops.partition import partition_row_indexing
+from kgcnn.ops.segment import segment_ops_by_name, segment_softmax
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingLocalEdges')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdges')
 class PoolingLocalEdges(GraphBaseLayer):
     """Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
     
     If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
     Apply e.g. segment_mean for index[0] incoming nodes. 
-    Important: edge_index[:,0] are sorted for segment-operation.
+    Important: tensor_index[:,0] are sorted for segment-operation.
     
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -32,26 +31,26 @@ class PoolingLocalEdges(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (list): of [node, edges, edge_index]
+            inputs (list): of [node, edges, tensor_index]
 
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - edges (tf.ragged): Edge or message features of shape (batch, [M], F)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], 2)
+                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
+                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
     
         Returns:
-            features: Pooled feature tensor of pooled edge features for each node.
+            tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node.
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 3)
+        dyn_inputs = inputs
         # We cast to values here
         nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
         edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
 
-        shiftind = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edgeind, node_part, edge_part,
-                                                                          partition_type_node="row_splits",
-                                                                          partition_type_edge="row_length",
-                                                                          to_indexing='batch',
-                                                                          from_indexing=self.node_indexing)
+        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
+                                          partition_type_target="row_splits",
+                                          partition_type_index="row_length",
+                                          to_indexing='batch',
+                                          from_indexing=self.node_indexing)
 
         nodind = shiftind[:, 0]  # Pick first index eg. ingoing
         dens = edge
@@ -61,11 +60,12 @@ class PoolingLocalEdges(GraphBaseLayer):
             nodind = tf.gather(nodind, node_order, axis=0)
             dens = tf.gather(dens, node_order, axis=0)
         # Pooling via e.g. segment_sum
-        out = kgcnn_ops_segment_operation_by_name(self.pooling_method, dens, nodind)
+        out = segment_ops_by_name(self.pooling_method, dens, nodind)
         if self.has_unconnected:
-            out = kgcnn_ops_scatter_segment_tensor_nd(out, nodind, tf.shape(nod))
+            out = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(out)[0]), axis=-1), out,
+                                tf.concat([tf.shape(nod)[:1], tf.shape(out)[1:]], axis=0))
 
-        out = self._kgcnn_map_output_ragged([out, node_part], "row_splits", 0)
+        out = tf.RaggedTensor.from_row_splits(out, node_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -78,14 +78,13 @@ class PoolingLocalEdges(GraphBaseLayer):
 PoolingLocalMessages = PoolingLocalEdges  # For now they are synonyms
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingWeightedLocalEdges')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedLocalEdges')
 class PoolingWeightedLocalEdges(GraphBaseLayer):
-    """
-    Pooling all edges or message/edge-like features per node, corresponding to node assigned by edge_indices.
+    """Pooling all edges or message/edge-like features per node, corresponding to node assigned by edge_indices.
     
     If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
     Apply e.g. segment_mean for index[0] incoming nodes. 
-    Important: edge_index[:,0] could be sorted for segment-operation.
+    Important: tensor_index[:,0] could be sorted for segment-operation.
     
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -108,28 +107,27 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (list): of [node, edges, edge_index, weights]
+            inputs (list): of [node, edges, tensor_index, weights]
 
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - edges (tf.ragged): Edge or message features of shape (batch, [M], F)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], 2)
-            - weights (tf.ragged): Edge or message weights. Must broadcast to edges or messages, e.g. (batch, [M], 1)
+                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
+                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
+                - weights (tf.RaggedTensor): Edge or message weights. Must broadcast to edges or messages,
+                  e.g. (batch, [M], 1)
 
         Returns:
-            features: Pooled feature tensor of pooled edge features for each node of shape (batch, [N], F)
+            tf.RaggedTensor: Pooled feature tensor of pooled edge features for each node of shape (batch, [N], F)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 4)
+        dyn_inputs = inputs
         # We cast to values here
         nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
         edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
         weights, _ = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
 
-        shiftind = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edgeind, node_part, edge_part,
-                                                                          partition_type_node="row_splits",
-                                                                          partition_type_edge="row_length",
-                                                                          to_indexing='batch',
-                                                                          from_indexing=self.node_indexing)
+        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
+                                          partition_type_target="row_splits", partition_type_index="row_length",
+                                          to_indexing='batch', from_indexing=self.node_indexing)
 
         wval = weights
         dens = edge * wval
@@ -143,15 +141,16 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
             wval = tf.gather(wval, node_order, axis=0)
 
         # Pooling via e.g. segment_sum
-        get = kgcnn_ops_segment_operation_by_name(self.pooling_method, dens, nodind)
+        get = segment_ops_by_name(self.pooling_method, dens, nodind)
 
         if self.normalize_by_weights:
             get = tf.math.divide_no_nan(get, tf.math.segment_sum(wval, nodind))  # +tf.eps
 
         if self.has_unconnected:
-            get = kgcnn_ops_scatter_segment_tensor_nd(get, nodind, tf.shape(nod))
+            get = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
+                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))
 
-        out = self._kgcnn_map_output_ragged([get, node_part], "row_splits", 0)
+        out = tf.RaggedTensor.from_row_splits(get, node_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -164,10 +163,9 @@ class PoolingWeightedLocalEdges(GraphBaseLayer):
 PoolingWeightedLocalMessages = PoolingWeightedLocalEdges  # For now they are synonyms
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingNodes')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodes')
 class PoolingNodes(GraphBaseLayer):
-    """
-    Polling all nodes per batch. The batch assignment is given by a length-tensor.
+    """Polling all nodes per batch. The batch assignment is given by a length-tensor.
     
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -186,16 +184,16 @@ class PoolingNodes(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (tf.ragged): Node features of shape (batch, [N], F)
+            inputs (tf.RaggedTensor): Node features of shape (batch, [N], F)
     
         Returns:
-            nodes (tf.tensor): Pooled node features of shape (batch, F)
+            tf.Tensor: Pooled node features of shape (batch, F)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged([inputs], 1)
+        dyn_inputs = [inputs]
         # We cast to values here
         nod, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
 
-        out = kgcnn_ops_segment_operation_by_name(self.pooling_method, nod, batchi)
+        out = segment_ops_by_name(self.pooling_method, nod, batchi)
 
         # Output should have correct shape
         return out
@@ -207,10 +205,9 @@ class PoolingNodes(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingWeightedNodes')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingWeightedNodes')
 class PoolingWeightedNodes(GraphBaseLayer):
-    """
-    Polling all nodes per batch. The batch assignment is given by a length-tensor.
+    """Polling all nodes per batch. The batch assignment is given by a length-tensor.
 
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -231,19 +228,19 @@ class PoolingWeightedNodes(GraphBaseLayer):
         Args:
             inputs (list): of [node, weights]
 
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - weights (tf.ragged): Node or message weights. Most broadcast to nodes. Shape (batch, [N], 1).
+                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
+                - weights (tf.RaggedTensor): Node or message weights. Most broadcast to nodes. Shape (batch, [N], 1).
 
         Returns:
-            nodes (tf.tensor): Pooled node features of shape (batch,F)
+            tf.Tensor: Pooled node features of shape (batch, F)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 2)
+        dyn_inputs = inputs
         # We cast to values here
         nod, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
         weights, _ = dyn_inputs[1].values, dyn_inputs[1].value_rowids()
 
         nod = tf.math.multiply(nod, weights)
-        out = kgcnn_ops_segment_operation_by_name(self.pooling_method, nod, batchi)
+        out = segment_ops_by_name(self.pooling_method, nod, batchi)
         # Output should have correct shape
         return out
 
@@ -254,10 +251,9 @@ class PoolingWeightedNodes(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingGlobalEdges')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingGlobalEdges')
 class PoolingGlobalEdges(GraphBaseLayer):
-    """
-    Pooling all edges per graph. The batch assignment is given by a length-tensor.
+    """Pooling all edges per graph. The batch assignment is given by a length-tensor.
 
     Args:
         pooling_method (str): Pooling method to use i.e. segment_function. Default is 'mean'.
@@ -276,16 +272,16 @@ class PoolingGlobalEdges(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (tf.ragged): Edge features or message embeddings of shape (batch, [M], F)
+            inputs (tf.RaggedTensor): Edge features or message embeddings of shape (batch, [M], F)
     
         Returns:
-            tf.tensor: Pooled edges feature list of shape (batch, F).
+            tf.Tensor: Pooled edges feature list of shape (batch, F).
         """
-        dyn_inputs = self._kgcnn_map_input_ragged([inputs], 1)
+        dyn_inputs = [inputs]
         # We cast to values here
         edge, batchi = dyn_inputs[0].values, dyn_inputs[0].value_rowids()
 
-        out = kgcnn_ops_segment_operation_by_name(self.pooling_method, edge, batchi)
+        out = segment_ops_by_name(self.pooling_method, edge, batchi)
         # Output already has correct shape and type
         return out
 
@@ -296,7 +292,7 @@ class PoolingGlobalEdges(GraphBaseLayer):
         return config
 
 
-@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingLocalEdgesLSTM')
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesLSTM')
 class PoolingLocalEdgesLSTM(GraphBaseLayer):
     """
     Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
@@ -304,7 +300,7 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
 
     If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
     Apply e.g. segment_mean for index[0] incoming nodes.
-    Important: edge_index[:,0] are sorted for segment-operation.
+    Important: tensor_index[:,0] are sorted for segment-operation.
 
     Args:
         units (int): Units for LSTM cell.
@@ -408,27 +404,26 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
         """Forward pass.
 
         Args:
-            inputs (list): [nodes, edges, edge_index]
-
-            - nodes (tf.ragged): Node features of shape (batch, [N], F)
-            - edges (tf.ragged): Edge or message features of shape (batch, [M], F)
-            - edge_index (tf.ragged): Edge indices of shape (batch, [M], 2)
+            inputs (list): [nodes, edges, tensor_index]
 
+                - nodes (tf.RaggedTensor): Node features of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message features of shape (batch, [M], F)
+                - tensor_index (tf.RaggedTensor): Edge indices of shape (batch, [M], 2)
 
         Returns:
-            features: Feature tensor of pooled edge features for each node of shape (batch, [N], F)
+            tf.RaggedTensor: Feature tensor of pooled edge features for each node of shape (batch, [N], F)
         """
-        dyn_inputs = self._kgcnn_map_input_ragged(inputs, 3)
+        dyn_inputs = inputs
         # We cast to values here
         nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_splits
         edge, _ = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
         edgeind, edge_part = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
 
-        shiftind = kgcnn_ops_change_edge_tensor_indexing_by_row_partition(edgeind, node_part, edge_part,
-                                                                          partition_type_node="row_splits",
-                                                                          partition_type_edge="row_length",
-                                                                          to_indexing='batch',
-                                                                          from_indexing=self.node_indexing)
+        shiftind = partition_row_indexing(edgeind, node_part, edge_part,
+                                          partition_type_target="row_splits",
+                                          partition_type_index="row_length",
+                                          to_indexing='batch',
+                                          from_indexing=self.node_indexing)
 
         nodind = shiftind[:, 0]  # Pick first index eg. ingoing
         dens = edge
@@ -446,9 +441,10 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
         if self.has_unconnected:
             # Need to fill tensor since the maximum node may not be also in pooled
             # Does not happen if all nodes are also connected
-            get = kgcnn_ops_scatter_segment_tensor_nd(get, nodind, tf.shape(nod))
+            get = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
+                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))
 
-        out = self._kgcnn_map_output_ragged([get, node_part], "row_splits", 0)
+        out = tf.RaggedTensor.from_row_splits(get, node_part, validate=self.ragged_validate)
         return out
 
     def get_config(self):
@@ -467,3 +463,130 @@ class PoolingLocalEdgesLSTM(GraphBaseLayer):
 
 
 PoolingLocalMessagesLSTM = PoolingLocalEdgesLSTM  # For now they are synonyms
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingLocalEdgesAttention')
+class PoolingLocalEdgesAttention(GraphBaseLayer):
+    r"""Pooling all edges or edge-like features per node, corresponding to node assigned by edge indices.
+    Uses attention for pooling. i.e. :math:`n_i =  \sum_j \alpha_{ij} e_{ij}`
+    The attention is computed via: :math:`\alpha_ij = \text{softmax}_j (a_{ij})` from the
+    attention coefficients :math:`a_{ij}`.
+    The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W n_i || W n_j)` and
+    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
+    In summary, :math:`n_i = \sum_j \text{softmax}_j (a_{ij}) e_{ij}` is computed by the layer.
+
+    An edge is defined by index tuple (i,j) with i<-j connection.
+    If graphs indices were in 'batch' mode, the layer's 'node_indexing' must be set to 'batch'.
+    Important: tensor_index[:,0] are sorted for segment-operation for pooling with respect to tensor_index[:,0].
+    """
+
+    def __init__(self, **kwargs):
+        """Initialize layer."""
+        super(PoolingLocalEdgesAttention, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(PoolingLocalEdgesAttention, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs: [node, edges, attention, edge_indices]
+
+                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
+                - edges (tf.RaggedTensor): Edge or message embeddings of shape (batch, [M], F)
+                - attention (tf.RaggedTensor): Attention coefficients of shape (batch, [M], 1)
+                - edge_indices (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], F)
+
+        Returns:
+            tf.RaggedTensor: Embedding tensor of pooled edge attentions for each node of shape (batch, [N], F)
+        """
+        dyn_inputs = inputs
+
+        # We cast to values here
+        nod, node_part = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
+        edge = dyn_inputs[1].values
+        attention = dyn_inputs[2].values
+        edgeind, edge_part = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
+
+        shiftind = partition_row_indexing(edgeind, node_part, edge_part, partition_type_target="row_length",
+                                          partition_type_index="row_length", to_indexing='batch',
+                                          from_indexing=self.node_indexing)
+
+        nodind = shiftind[:, 0]  # Pick first index eg. ingoing
+        dens = edge
+        ats = attention
+        if not self.is_sorted:
+            # Sort edgeindices
+            node_order = tf.argsort(nodind, axis=0, direction='ASCENDING', stable=True)
+            nodind = tf.gather(nodind, node_order, axis=0)
+            dens = tf.gather(dens, node_order, axis=0)
+            ats = tf.gather(ats, node_order, axis=0)
+
+        # Apply segmented softmax
+        ats = segment_softmax(ats, nodind)
+        get = dens * ats
+        get = tf.math.segment_sum(get, nodind)
+
+        if self.has_unconnected:
+            # Need to fill tensor since the maximum node may not be also in pooled
+            # Does not happen if all nodes are also connected
+            get = tf.scatter_nd(tf.keras.backend.expand_dims(tf.range(tf.shape(get)[0]), axis=-1), get,
+                                tf.concat([tf.shape(nod)[:1], tf.shape(get)[1:]], axis=0))
+
+        out = tf.RaggedTensor.from_row_lengths(get, node_part, validate=self.ragged_validate)
+        return out
+
+    def get_config(self):
+        """Update layer config."""
+        config = super(PoolingLocalEdgesAttention, self).get_config()
+        return config
+
+
+@tf.keras.utils.register_keras_serializable(package='kgcnn', name='PoolingNodesAttention')
+class PoolingNodesAttention(GraphBaseLayer):
+    r"""Pooling all nodes.
+    Uses attention for pooling. i.e. :math:`s =  \sum_j \alpha_{i} n_i`
+    The attention is computed via: :math:`\alpha_i = \text{softmax}_i(a_i)` from the attention coefficients :math:`a_i`.
+    The attention coefficients must be computed beforehand by edge features or by :math:`\sigma( W [s || n_i])` and
+    are passed to this layer as input. Thereby this layer has no weights and only does pooling.
+    In summary, :math:`s =  \sum_i \text{softmax}_j(a_i) n_i` is computed by the layer.
+
+    """
+
+    def __init__(self, **kwargs):
+        """Initialize layer."""
+        super(PoolingNodesAttention, self).__init__(**kwargs)
+
+    def build(self, input_shape):
+        """Build layer."""
+        super(PoolingNodesAttention, self).build(input_shape)
+
+    def call(self, inputs, **kwargs):
+        """Forward pass.
+
+        Args:
+            inputs: [nodes, attention]
+
+                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
+                - attention (tf.RaggedTensor): Attention coefficients of shape (batch, [N], 1)
+
+        Returns:
+            tf.Tensor: Embedding tensor of pooled node of shape (batch, F)
+        """
+        dyn_inputs = inputs
+        # We cast to values here
+        nod, batchi, target_len = dyn_inputs[0].values, dyn_inputs[0].value_rowids(), dyn_inputs[0].row_lengths()
+        ats = dyn_inputs[1].values
+
+        ats = segment_softmax(ats, batchi)
+        get = nod * ats
+        out = tf.math.segment_sum(get, batchi)
+
+        return out
+
+    def get_config(self):
+        """Update layer config."""
+        config = super(PoolingNodesAttention, self).get_config()
+        return config
