@@ -7,16 +7,17 @@ from scipy.optimize import brentq
 
 @tf.function
 def tf_spherical_bessel_jn_explicit(x, n=0):
-    r"""Compute spherical bessel functions $j_n(x)$ for constant positive integer $n$ explicitly.
-    TensorFlow has to cache the function for each n. No gradient through n or very large number of n's is possible.
+    r"""Compute spherical bessel functions :math:`j_n(x)` for constant positive integer :math:`n` explicitly.
+    TensorFlow has to cache the function for each :math:`n`. No gradient through :math:`n` or very large number
+    of :math:`n`'s is possible.
     Source: https://dlmf.nist.gov/10.49
 
     Args:
-        x (tf.tensor): Values to compute jn(x) for.
-        n (int): Positive integer for the bessel order n.
+        x (tf.Tensor): Values to compute :math:`j_n(x)` for.
+        n (int): Positive integer for the bessel order :math:`n`.
 
     Returns:
-        tf.tensor: Spherical bessel function of order n
+        tf.Tensor: Spherical bessel function of order :math:`n`
     """
     sin_x = tf.sin(x - n * np.pi / 2)
     cos_x = tf.cos(x - n * np.pi / 2)
@@ -37,14 +38,14 @@ def tf_spherical_bessel_jn_explicit(x, n=0):
 
 @tf.function
 def tf_spherical_bessel_jn(x, n=0):
-    """Compute spherical bessel functions jn(x) for constant positive integer n via recursion.
+    r"""Compute spherical bessel functions :math:`j_n(x)` for constant positive integer :math:`n` via recursion.
 
     Args:
-        x (tf.tensor): Values to compute jn(x) for.
-        n (int): Positive integer for the bessel order n.
+        x (tf.Tensor): Values to compute :math:`j_n(x)` for.
+        n (int): Positive integer for the bessel order :math:`n`.
 
     Returns:
-        tf.tensor: Spherical bessel function of order n
+        tf.tensor: Spherical bessel function of order :math:`n`
     """
     if n < 0:
         raise ValueError("Order parameter must be >= 0 for this implementation of spherical bessel function.")
@@ -64,14 +65,16 @@ def tf_spherical_bessel_jn(x, n=0):
 
 @tf.function
 def tf_legendre_polynomial_pn(x, n=0):
-    """Compute the (non-associated) Legendre polynomial for constant positive integer n via explicit formula.
+    r"""Compute the (non-associated) Legendre polynomial :math:`P_n(x)` for constant positive integer :math:`n`
+    via explicit formula.
+    Closed form: https://en.wikipedia.org/wiki/Legendre_polynomials
 
     Args:
-        x (tf.tensor): Values to compute Pn(x) for.
-        n (int): Positive integer for n in Pn(x).
+        x (tf.Tensor): Values to compute :math:`P_n(x)` for.
+        n (int): Positive integer for :math:`n` in :math:`P_n(x)`.
 
     Returns:
-        tf.tensor: Legendre Polynomial of order n.
+        tf.tensor: Legendre Polynomial of order :math:`n`.
     """
     out_sum = tf.zeros_like(x)
     prefactors = [
@@ -85,14 +88,14 @@ def tf_legendre_polynomial_pn(x, n=0):
 
 @tf.function
 def tf_spherical_harmonics_yl(theta, l=0):
-    """Compute the spherical harmonics for m=0 and constant non-integer l.
+    r"""Compute the spherical harmonics :math:`Y_{ml}(\cos\theta)` for :math:`m=0` and constant non-integer :math:`l`.
 
     Args:
-        theta (tf.tensor): Values to compute Yl(theta) for.
-        l (int): Positive integer for l in Yl(x).
+        theta (tf.Tensor): Values to compute :math:`Y_l(\cos\theta)` for.
+        l (int): Positive integer for :math:`l` in :math:`Y_l(\cos\theta)`.
 
     Returns:
-        tf.tensor: Spherical harmonics for m=0 and constant non-integer l.
+        tf.tensor: Spherical harmonics for :math:`m=0` and constant non-integer :math:`l`.
     """
     x = tf.cos(theta)
     out_sum = tf.zeros_like(x)
@@ -106,40 +109,55 @@ def tf_spherical_harmonics_yl(theta, l=0):
     return out_sum
 
 
-# @tf.function
-# def tf_associated_legendre_polynomial(x, l=0, m=0):
-#     """Compute the associated Legendre polynomial for m and constant positive integer l via explicit formula.
-#
-#     Args:
-#         x (tf.tensor): Values to compute Plm(x) for.
-#         l (int): Positive integer for l in Plm(x).
-#         m (int): Positive/Negative integer for m in Plm(x).
-#
-#     Returns:
-#         tf.tensor: Legendre Polynomial of order n.
-#     """
-#     if m==0:
-#     else:
-#         prefactors = [ for k in range(m,l)]
-#         powers = [ for k in range(m,l)]
-#         if
+@tf.function
+def tf_associated_legendre_polynomial(x, l=0, m=0):
+    """Compute the associated Legendre polynomial :math:`P_{l}^{m}(x)` for :math:`m` and constant positive
+    integer :math:`l` via explicit formula.
+    Closed Form from: https://en.wikipedia.org/wiki/Associated_Legendre_polynomials
+
+    Args:
+        x (tf.Tensor): Values to compute :math:`P_{l}^{m}(x)` for.
+        l (int): Positive integer for :math:`l` in :math:`P_{l}^{m}(x)`.
+        m (int): Positive/Negative integer for :math:`m` in :math:`P_{l}^{m}(x)`.
+
+    Returns:
+        tf.tensor: Legendre Polynomial of order n.
+    """
+    if np.abs(m)>l:
+        raise ValueError("Error: Legendre polynomial must have -l<= m <= l")
+    if l<0:
+        raise ValueError("Error: Legendre polynomial must have l>=0")
+    if m < 0:
+        m = -m
+        neg_m = float(np.power(-1,m) * sp.special.factorial(l-m)/sp.special.factorial(l+m))
+    else:
+        neg_m = 1
+
+    x_prefactor = tf.pow(1 - tf.square(x), m/2) * float(np.power(-1,m) * np.power(2,l))
+    sum_out = tf.zeros_like(x)
+    for k in range(m, l+1):
+        sum_out += tf.pow(x, k-m) * float(sp.special.factorial(k)/sp.special.factorial(k-m)*sp.special.binom(l,k)*
+                                         sp.special.binom((l+k-1)/2,l))
+
+    return sum_out*x_prefactor*neg_m
 
 
 def spherical_bessel_jn(r, n):
-    """Compute spherical Bessel function jn(r) via scipy.
+    r"""Compute spherical Bessel function :math:`j_n(r)` via scipy.
 
     Args:
-        r (array_like): Argument
-        n (array_like): Order.
+        r (np.ndarray): Argument
+        n (np.ndarray): Order.
 
     Returns:
-        array_like: Values of the spherical Bessel function
+        np.array: Values of the spherical Bessel function
     """
     return np.sqrt(np.pi / (2 * r)) * sp.special.jv(n + 0.5, r)
 
 
 def spherical_bessel_jn_zeros(n, k):
-    """Compute the first k zeros of the spherical bessel functions up to order n (excluded).
+    r"""Compute the first :math:`k` zeros of the spherical bessel functions :math:`j_n(r)` up to
+    order :math:`n` (excluded).
     Taken from https://github.com/klicperajo/dimenet.
 
     Args:
@@ -147,7 +165,7 @@ def spherical_bessel_jn_zeros(n, k):
         k: Number of zero crossings.
 
     Returns:
-        zeros_jn: List of zero crossings of shape (n,k)
+        np.ndarray: List of zero crossings of shape (n, k)
     """
     zerosj = np.zeros((n, k), dtype="float32")
     zerosj[0] = np.arange(1, k + 1) * np.pi
@@ -164,8 +182,8 @@ def spherical_bessel_jn_zeros(n, k):
 
 
 def spherical_bessel_jn_normalization_prefactor(n, k):
-    """Compute the normalization or rescaling pre-factor for the spherical bessel functions up to
-    order n (excluded) and maximum frequency k (excluded).
+    r"""Compute the normalization or rescaling pre-factor for the spherical bessel functions :math:`j_n(r)` up to
+    order :math:`n` (excluded) and maximum frequency :math:`k` (excluded).
     Taken from https://github.com/klicperajo/dimenet.
 
     Args:
@@ -173,7 +191,7 @@ def spherical_bessel_jn_normalization_prefactor(n, k):
         k: frequency.
 
     Returns:
-        norm: Normalization of shape (n, k)
+        np.ndarray: Normalization of shape (n, k)
     """
     zeros = spherical_bessel_jn_zeros(n, k)
     normalizer = []
